package com.wizarpos.mobileemv.ui

import android.Manifest
import android.app.PendingIntent
import android.content.Context
import android.content.ContextWrapper
import android.content.Intent
import android.content.IntentFilter
import android.content.pm.PackageManager
import android.graphics.Bitmap
import android.location.LocationManager
import android.nfc.NfcAdapter
import android.nfc.Tag
import android.nfc.tech.IsoDep
import android.nfc.tech.MifareClassic
import android.nfc.tech.NfcA
import android.os.Build
import android.os.Bundle
import android.provider.Settings
import android.telephony.TelephonyManager
import android.text.*
import android.text.method.ScrollingMovementMethod
import android.util.Log
import android.view.View
import android.view.ViewGroup
import android.widget.TextView
import android.widget.Toast
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import androidx.recyclerview.widget.LinearLayoutManager
import androidx.recyclerview.widget.RecyclerView
import com.cloudpos.jniinterface.EMVJNIInterfaceAir
import com.cloudpos.jniinterface.EMVJNIInterfaceAir.emv_get_tag_data
import com.cloudpos.jniinterface.EMVJNIInterfaceAir.emv_is_tag_present
import com.cloudpos.jniinterface.IFuntionListener
import com.orhanobut.hawk.Hawk
import com.qmuiteam.qmui.arch.QMUIActivity
import com.qmuiteam.qmui.util.QMUIDisplayHelper
import com.qmuiteam.qmui.util.QMUIStatusBarHelper
import com.qmuiteam.qmui.util.QMUIViewHelper
import com.qmuiteam.qmui.widget.dialog.QMUIDialog
import com.qmuiteam.qmui.widget.dialog.QMUITipDialog
import com.qmuiteam.qmui.widget.popup.QMUIPopup
import com.qmuiteam.qmui.widget.popup.QMUIPopups
import com.qmuiteam.qmui.widget.section.QMUISection
import com.qmuiteam.qmui.widget.section.QMUIStickySectionAdapter
import com.wizarpos.libemvair.emv.EmvProcess
import com.wizarpos.util.AppUtil
import com.wizarpos.util.ByteUtil
import com.wizarpos.util.NumberUtil
import com.wizarpos.util.StringUtil
import com.wizarpos.mobileemv.App
import com.wizarpos.mobileemv.R
import com.wizarpos.mobileemv.ui.presenter.HomeViewPagerAdapter
import com.wizarpos.mobileemv.ui.presenter.QueryTransAdapter
import com.wizarpos.mobileemv.ui.presenter.QueryTransSectionHeader
import com.wizarpos.mobileemv.ui.presenter.QueryTransSectionItem
import kotlinx.android.synthetic.main.activity_main.*
import kotlinx.android.synthetic.main.dialog_image.*
import kotlinx.android.synthetic.main.progress_msg.*
import kotlinx.android.synthetic.main.view_pager_log.view.*
import kotlinx.android.synthetic.main.view_pager_record.view.*
import kotlinx.coroutines.*
import java.io.ByteArrayOutputStream

class MainActivity : QMUIActivity(), IUiUpdate, IFuntionListener {

	private val TAG = "MainActivity"
	private val REQUEST_CODE_OPEN_GPS = 1
	private val REQUEST_CODE_PERMISSION_LOCATION = 2
	private val REQUEST_CODE_READ_PHONE_STATE = 3


	private lateinit var mNormalPopup: QMUIPopup
	private lateinit var mNormalPopupText: TextView
	private var mTipDialog: QMUITipDialog? = null
	private var mProgressDialog: QMUIDialog? = null
	private var mToast: Toast? = null
	private var mViewMap: HashMap<Int, View> = HashMap()
	private val mRecordAdapter = QueryTransAdapter()

	private lateinit var mTextLog: TextView
	private var mNfcAdapter: NfcAdapter? = null
	private lateinit var mPendingIntent: PendingIntent
	private lateinit var mFilters: Array<IntentFilter>
	private lateinit var mTechLists: Array<Array<String>>

	var uiUpdate: IUiUpdate? = null
	private val emvProcess = EmvProcess.getInstance()
	private var isActive = false

	override fun onCreate(savedInstanceState: Bundle?) {
		super.onCreate(savedInstanceState)
		QMUIStatusBarHelper.translucent(this)
		setContentView(R.layout.activity_main)
		initViews()

		checkPermissions()

		mNfcAdapter = NfcAdapter.getDefaultAdapter(this)

		mPendingIntent = PendingIntent.getActivity(this, 0, intent, PendingIntent.FLAG_UPDATE_CURRENT or PendingIntent.FLAG_MUTABLE)
		try {
			mFilters = arrayOf(IntentFilter(NfcAdapter.ACTION_TECH_DISCOVERED, "*/*"), IntentFilter(NfcAdapter.ACTION_TAG_DISCOVERED))

		} catch (e: Exception) {
			e.printStackTrace()
		}
		mTechLists = arrayOf(arrayOf(NfcA::class.java.name), arrayOf(IsoDep::class.java.name))

		Log.d(TAG, "onCreate, intent.act:${intent.action}")
		/**
		 * when nfc discover card, use nfc processor for emv processing
		 */
		if (intent.action == NfcAdapter.ACTION_TECH_DISCOVERED) {
			nfcTagProcess(intent.getParcelableExtra<Tag>(NfcAdapter.EXTRA_TAG) as Tag)
		}
	}

	override fun onNewIntent(intent: Intent?) {
		super.onNewIntent(intent)
		setIntent(intent)
		Log.d(TAG, "onNewIntent, intent.act:${intent?.action}")
		uiUpdate = this
		//process nfc tag
		val tag: Tag? = intent?.getParcelableExtra(NfcAdapter.EXTRA_TAG)
		if (tag != null) {
			nfcTagProcess(tag)
		}
	}


	override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<String>, grantResults: IntArray) {
		super.onRequestPermissionsResult(requestCode, permissions, grantResults)
		Log.i(TAG, "onRequestPermissionsResult:$requestCode")
		when (requestCode) {
			REQUEST_CODE_READ_PHONE_STATE,
			REQUEST_CODE_PERMISSION_LOCATION -> if (grantResults.isNotEmpty()) {
				for (i in grantResults.indices) {
					if (grantResults[i] == PackageManager.PERMISSION_GRANTED) {
						onPermissionGranted(permissions[i])
					}
				}
			}
		}
	}

	override fun onResume() {
		super.onResume()
		mNfcAdapter?.also {
			if (it.isEnabled) {
				it.enableForegroundDispatch(this, mPendingIntent, mFilters, mTechLists)
			}
		}
	}

	override fun onPause() {
		super.onPause()
		mNfcAdapter?.also {
			if (it.isEnabled) {
				it.disableForegroundDispatch(this)
			}
		}

	}


	override fun addContentLog(log: String) {
		Log.i(TAG, "addContentLog:$log")
		GlobalScope.launch(Dispatchers.Main) {
			Log.i(TAG, "addContentLog in coroutine:$log")
			QMUIViewHelper.playBackgroundBlinkAnimation(mTextLog, ContextCompat.getColor(App.instance, R.color.bar_divider))
			mTextLog.append(log)
			mTextLog.append("\n\n")
		}

	}

	override fun clearContentLog() {
		GlobalScope.launch(Dispatchers.Main) {
			mTextLog.text = ""
		}
	}


	override fun getAmount(): String {
		return edit_amount_input.text.toString()
	}

	override fun showTipDialog(msg: String, iconType: Int, duration: Int) {
		GlobalScope.launch(Dispatchers.Main) {
			if (mTipDialog != null) {
				mTipDialog?.dismiss()
			}
			mTipDialog = QMUITipDialog.Builder(this@MainActivity)
				.setIconType(iconType)
				.setTipWord(msg)
				.create(duration > 0)
			mTipDialog?.show()

			if (duration > 0) {
				withContext(Dispatchers.IO) {
					delay(duration.toLong())
					withContext(Dispatchers.Main) {
						hideTipDialog()
					}
				}
			}

		}
	}


	override fun hideTipDialog() {
		GlobalScope.launch(Dispatchers.Main) {
			if (mTipDialog != null) {
				mTipDialog?.dismiss()
				mTipDialog = null
			}
		}

	}

	override fun showImageDialog(title: String, bitmap: Bitmap) {
		GlobalScope.launch(Dispatchers.Main) {
			if (mTipDialog != null) {
				mTipDialog?.dismiss()
			}
			val dialog = QMUIDialog.CustomDialogBuilder(this@MainActivity)
				.setLayout(R.layout.dialog_image)
				.addAction("CONFIRM") { dialog, index -> dialog.dismiss() }
				.setTitle(title)
				.create()
			dialog.img_fp.setImageBitmap(bitmap)
			dialog.show()
		}
	}

	override fun setUpPan(pan: String?, resourceId: Int, cardExpiryDate: String?) {
		GlobalScope.launch(Dispatchers.Main) {
			Log.i(TAG, "pan:$pan, expiryDate:$cardExpiryDate")
			text_card_pan.text = pan
			if (resourceId > 0) {
				image_card_org.visibility = View.VISIBLE
				image_card_org.setImageResource(resourceId)
				if (cardExpiryDate != null) text_expiry_date.text = cardExpiryDate
			}
		}
	}

	override fun showMessageDialog(title: String, msg: String, callback: IDialogCallback) {
		GlobalScope.launch(Dispatchers.Main) {
			QMUIDialog.MessageDialogBuilder(this@MainActivity)
				.setTitle(title)
				.setMessage(msg)
				.setCancelable(false)
				.setCanceledOnTouchOutside(false)
				.addAction("CANCEL") { dialog, _ ->
					callback.callBack(0, null)
					dialog.dismiss()
				}
				.addAction("CONFIRM") { dialog, _ ->
					callback.callBack(1, null)
					dialog.dismiss()
				}
				.create(R.style.QMUI_Dialog)
				.show()
		}

	}

	override fun showMenuDialog(title: String, menuList: Array<String>, callback: IDialogCallback) {
		GlobalScope.launch(Dispatchers.Main) {
			QMUIDialog.MenuDialogBuilder(this@MainActivity)
				.setTitle(title)
				.setCancelable(false)
				.setCanceledOnTouchOutside(false)
				.addItems(menuList) { dialog, which ->
					callback.callBack(which, null)
					dialog.dismiss()
				}
				.create(R.style.QMUI_Dialog)
				.show()
		}

	}

	override fun showProgressBarDialog(title: String, curProgress: Int, progressTotal: Int, updateProgress: Boolean, hide: Boolean) {
		GlobalScope.launch(Dispatchers.Main) {
			if (hide) {
				mProgressDialog?.dismiss()
				return@launch
			}

			if (updateProgress) {
				mProgressDialog?.let {
					it.progress_bar.progress = curProgress
				}
				return@launch
			}
			mProgressDialog = QMUIDialog.CustomDialogBuilder(this@MainActivity)
				.setLayout(R.layout.progress_msg)
				.setCanceledOnTouchOutside(false)
				.setCancelable(false)
				.setTitle(title)
				.create(R.style.QMUI_Dialog)

			val processBar = mProgressDialog!!.progress_bar
			processBar.maxValue = progressTotal
			processBar.progress = curProgress
			processBar.setQMUIProgressBarTextGenerator { progressBar, value, maxValue ->
				"${100 * value / maxValue}%"
			}

			mProgressDialog?.show()
		}
	}

	override fun setUpCardRecord(sectionList: ArrayList<QMUISection<QueryTransSectionHeader, QueryTransSectionItem>>) {
		GlobalScope.launch(Dispatchers.Main) {
			mRecordAdapter.setData(sectionList)
		}
	}

	override fun showPopup(str: String) {
//		mNormalPopupText.text = str
//		mNormalPopup.show(fab_start_transaction)
	}

	override fun setUpBtnStart(isStart: Boolean) {
//		GlobalScope.launch(Dispatchers.Main) {
//			fab_start_transaction.tag = isStart
//			if (isStart) {
//				fab_start_transaction.setImageResource(R.drawable.float_start)
//				fab_start_transaction.backgroundTintList = ColorStateList.valueOf(resources.getColor(R.color.app_color_theme_5))
//			} else {
//				fab_start_transaction.setImageResource(R.drawable.float_stop)
//				fab_start_transaction.backgroundTintList = ColorStateList.valueOf(resources.getColor(R.color.app_color_theme_1))
//			}
//		}
	}

	override fun initUI() {
		GlobalScope.launch(Dispatchers.Main) {
			mTextLog.text = ""
			text_card_pan.text = ""
			text_expiry_date.text = ""
			image_card_org.visibility = View.INVISIBLE
			mRecordAdapter.setData(null)
		}
	}

	override fun showToast(msg: String) {
		mToast?.cancel()
		mToast = Toast.makeText(this, msg, Toast.LENGTH_SHORT)
		mToast?.show()
	}

	private fun checkPermissions() {

		val permissions = mutableSetOf(
			Manifest.permission.ACCESS_FINE_LOCATION,
			Manifest.permission.READ_PHONE_STATE,
		)
		if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.S) {
			permissions.add(Manifest.permission.BLUETOOTH_CONNECT)
			permissions.add(Manifest.permission.BLUETOOTH_SCAN)
		}

		val permissionDeniedList = ArrayList<String>()
		for (permission in permissions) {
			val permissionCheck = ContextCompat.checkSelfPermission(this, permission)
			if (permissionCheck == PackageManager.PERMISSION_GRANTED) {
				onPermissionGranted(permission)
			} else {
				permissionDeniedList.add(permission)
			}
		}
		if (permissionDeniedList.isNotEmpty()) {
			val deniedPermissions = permissionDeniedList.toTypedArray()
			ActivityCompat.requestPermissions(this, deniedPermissions, REQUEST_CODE_PERMISSION_LOCATION)
		}
	}


	private fun onPermissionGranted(permission: String) {
		Log.i(TAG, "onPermissionGranted, $permission")
		when (permission) {
			Manifest.permission.READ_PHONE_STATE -> registerEmvkernel()
			Manifest.permission.ACCESS_FINE_LOCATION -> if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M && !checkGPSIsOpen()) {

//				QMUIDialog.MessageDialogBuilder(this)
//					.setTitle("Tip")
//					.setMessage("need open position function")
//					.setCanceledOnTouchOutside(false)
//					.setCancelable(false)
//					.addAction("Cancel") { dialog, _ ->
//						finish()
//						dialog.dismiss()
//					}
//					.addAction(0, "Go to settings", QMUIDialogAction.ACTION_PROP_NEGATIVE) { dialog, _ ->
//						val intent = Intent(Settings.ACTION_LOCATION_SOURCE_SETTINGS)
//						startActivityForResult(intent, REQUEST_CODE_OPEN_GPS)
//						dialog.dismiss()
//					}
//					.create(R.style.QMUI_Dialog)
//					.show()
			}
		}
	}

	private fun checkGPSIsOpen(): Boolean {
		val locationManager = this.getSystemService(Context.LOCATION_SERVICE) as LocationManager
		val result = locationManager.isProviderEnabled(LocationManager.GPS_PROVIDER)
		Log.i(TAG, "GPS_PROVIDER:$result")
		return result
	}


	private fun initViews() {

		setUpBtnStart(true)
//		fab_start_transaction.setOnClickListener {
//			if (mTransactionNfc.mIsoDep != null) {
//				showTipDialog("nfc is processing!", QMUITipDialog.Builder.ICON_TYPE_INFO, 1000)
//				return@setOnClickListener
//			}
//
//			if (fab_start_transaction.tag as Boolean) {
//				if (!isActive) {
//					showTipDialog("emvkernel was not active!", QMUITipDialog.Builder.ICON_TYPE_FAIL, 3000)
//				} else {
//					mTransactionMpos.startTransaction()
//				}
//			} else {
//				mTransactionMpos.cancelTransaction()
//			}
//		}


		mNormalPopupText = TextView(this)
		mNormalPopupText.setLineSpacing(QMUIDisplayHelper.dp2px(this, 4)
			                                .toFloat(), 1.0f)
		val padding = QMUIDisplayHelper.dp2px(this, 20)
		mNormalPopupText.setPadding(padding, padding, padding, padding)
		mNormalPopupText.setTextColor(ContextCompat.getColor(this, R.color.app_color_description))

		mNormalPopup = QMUIPopups.popup(this, QMUIDisplayHelper.dp2px(this, 250))
			.preferredDirection(QMUIPopup.DIRECTION_TOP)
			.view(mNormalPopupText)
			.animStyle(QMUIPopup.ANIM_GROW_FROM_CENTER)

		edit_amount_input.inputType = InputType.TYPE_NUMBER_FLAG_DECIMAL or InputType.TYPE_CLASS_NUMBER
		edit_amount_input.filters = arrayOf<InputFilter>(object : InputFilter {
			var decimalNumber = 2

			override fun filter(source: CharSequence, start: Int, end: Int, dest: Spanned, dstart: Int, dend: Int): CharSequence? {
				val sourceContent = source.toString()
				val lastInputContent = dest.toString()

				if (TextUtils.isEmpty(sourceContent)) {
					return ""
				}
				if (sourceContent == "." && lastInputContent.isEmpty()) {
					return "0."
				}
				if (sourceContent == "0" && lastInputContent.isEmpty()) {
					return "0."
				}
				if (lastInputContent.contains(".")) {
					val index = lastInputContent.indexOf(".")
					if (dend - index >= decimalNumber + 1) {
						return ""
					}

				}
				return null
			}
		})

		edit_amount_input.addTextChangedListener(object : TextWatcher {
			lateinit var temp: String

			override fun beforeTextChanged(s: CharSequence, start: Int, count: Int, after: Int) {}

			override fun onTextChanged(s: CharSequence, start: Int, before: Int, count: Int) {}

			override fun afterTextChanged(s: Editable) {
				temp = s.toString()
				val posDot = temp.indexOf(".")
				//没有小数点，控制小数点前的位数
				if (posDot <= 0) {
					if (temp.length <= 7) {
						return
					} else {
						//删除第二位后面的内容
						s.delete(7, temp.length)
						return
					}
				}
			}
		})


		val tabBuilder = tabs.tabBuilder()

		tabs.addTab(tabBuilder.setText("TRANSACTION LOG")
			            .build(this))
			.addTab(tabBuilder.setText("CARD RECORDS")
				        .build(this))


		mViewMap[0] = layoutInflater.inflate(R.layout.view_pager_log, pager, false)
		mViewMap[1] = layoutInflater.inflate(R.layout.view_pager_record, pager, false)

		val recordSectionLayout = (mViewMap[1] as View).pager_record_section_layout
		recordSectionLayout.setLayoutManager(object : LinearLayoutManager(this) {
			override fun generateDefaultLayoutParams(): RecyclerView.LayoutParams {
				return RecyclerView.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT, ViewGroup.LayoutParams.WRAP_CONTENT)
			}
		})

		mRecordAdapter.setCallback(object : QMUIStickySectionAdapter.Callback<QueryTransSectionHeader, QueryTransSectionItem> {
			override fun loadMore(section: QMUISection<QueryTransSectionHeader, QueryTransSectionItem>?, loadMoreBefore: Boolean) {
			}

			override fun onItemLongClick(holder: QMUIStickySectionAdapter.ViewHolder?, position: Int): Boolean {
				return true
			}

			override fun onItemClick(holder: QMUIStickySectionAdapter.ViewHolder?, position: Int) {
				if (mRecordAdapter.getItemViewType(position) == QMUIStickySectionAdapter.ITEM_TYPE_SECTION_HEADER) {
					mRecordAdapter.toggleFold(position, false)
				}
			}

		})

		recordSectionLayout.setAdapter(mRecordAdapter, true)
		pager.adapter = HomeViewPagerAdapter(this, mViewMap)
		tabs.setupWithViewPager(pager, false)


		mTextLog = (mViewMap[0] as View).pager_log_text
		mTextLog.movementMethod = ScrollingMovementMethod.getInstance()
	}

	fun cardFound() {
		EMVJNIInterfaceAir.registerFunctionListener(this)
		emvProcess.setCardProcessor(App.instance.isoDep)
		GlobalScope.launch(Dispatchers.IO) {
			Log.i(TAG, "cardFound")
			uiUpdate?.initUI()
			try {
				App.instance.isoDep?.connect()
			} catch (e: Exception) {
				Log.e(TAG, "isoDep connect exception")
				e.printStackTrace()
				uiUpdate?.showTipDialog(e.localizedMessage, QMUITipDialog.Builder.ICON_TYPE_FAIL, 2000)
				return@launch
			}

			val data = Bundle()
			data.putByte(EmvProcess.PARAM_CARD_TYPE, EMVJNIInterfaceAir.CARD_TYPE_CONTACTLESS)
			uiUpdate?.addContentLog("###### start emv process, read record ######")
			emvProcess.start("0", EMVJNIInterfaceAir.TRANS_GOODS_SERVICE, data)
		}
	}

	private fun nfcTagProcess(tag: Tag) {
		val mifare = MifareClassic.get(tag)
		if (mifare != null) {
			clearContentLog()
			addContentLog("MifareClassic found\n size:${mifare.size} sectors:${mifare.sectorCount} blocks:${mifare.blockCount}")
			GlobalScope.launch(Dispatchers.IO) {
				testMifare(mifare)
			}

		} else {
			App.instance.isoDep = IsoDep.get(tag)
			cardFound()
		}
	}

	private suspend fun testMifare(mifare: MifareClassic) {
		try {
			mifare.connect()
			var delay = 0
			withContext(Dispatchers.Main) {
				val delayStr = edit_mifare_delay.text?.toString()
				if (!delayStr.isNullOrEmpty()) {
					delay = delayStr.toInt()
				}
			}
			for (i in 0 until 1) {
				var authenticated = true
				if (mifare.authenticateSectorWithKeyA(i, MifareClassic.KEY_MIFARE_APPLICATION_DIRECTORY)) {
					addContentLog("authenticated sector $i with KEY_MIFARE_APPLICATION_DIRECTORY")
				} else if (mifare.authenticateSectorWithKeyA(i, MifareClassic.KEY_DEFAULT)) {
					addContentLog("authenticated sector $i with KEY_DEFAULT")
				} else if (mifare.authenticateSectorWithKeyA(i, MifareClassic.KEY_NFC_FORUM)) {
					addContentLog("authenticated sector $i with KEY_NFC_FORUM")
				} else {
					authenticated = false
					addContentLog("authenticated sector $i failed")
				}

				if (authenticated) {
					for (j in 0 until 4) {
						delay(delay.toLong())
						val blockIdx = i * 4 +j
						val blockData = mifare.readBlock(blockIdx)
						addContentLog("read block $blockIdx: ${ByteUtil.arrayToHexStr(blockData)}")
					}
				}
			}
		}catch (e: Exception) {
			e.printStackTrace()
			withContext(Dispatchers.Main) {
				showToast("read block exception:\n ${e.message}")
			}
		}
	}

	private fun registerEmvkernel() {
		Log.i(TAG, "register Emv kernel")
		EMVJNIInterfaceAir.emv_kernel_initialize()
		EMVJNIInterfaceAir.registerFunctionListener(this)
		val telephonyManager = this.getSystemService(Context.TELEPHONY_SERVICE) as TelephonyManager
		val imei = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
//				telephonyManager.imei
			Settings.Secure.getString(contentResolver, Settings.Secure.ANDROID_ID)
		} else {
			Build.SERIAL
		}
		GlobalScope.launch(Dispatchers.IO) {
//			withContext(Dispatchers.Main) {
//				showTipDialog("registering..", QMUITipDialog.Builder.ICON_TYPE_LOADING, -1)
//			}
//			var license = if (Hawk.contains("emv_license")) {
//				Hawk.get("emv_license")
//			} else {
//				EmvProcess.getInstance()
//					.registOnlineLicense("86AT240611052", imei) //安可信 LICENSE ID 86AT240611052
//					.also {
//						Hawk.put("emv_license", it)
//					}
//			}

			var activeRet = EmvProcess.getInstance()
				.activeEmvkernel(null)
			Log.i(TAG, "1st active emvkernel:$activeRet")

//			if (activeRet == -2) {//expired
//				license = EmvProcess.getInstance()
//					.updateOnlineLicense(license)
//				if (license != null) {
//					Hawk.put("emv_license", license)
//					activeRet = EmvProcess.getInstance()
//						.activeEmvkernel(license)
//					Log.i(TAG, "2nd active emvkernel:$activeRet")
//				}
//			}
			this@MainActivity.isActive = (activeRet >= 0)
			withContext(Dispatchers.Main) {
				val msg = if (activeRet >= 0) "register success" else "register failed:$activeRet"
				showTipDialog(msg, QMUITipDialog.Builder.ICON_TYPE_INFO, 2000)
			}
		}
	}

	private fun readLine(assetName: String): ByteArray? {
		val buffer = ByteArrayOutputStream()
		try {
			val contextWrapper = ContextWrapper(this)
			val assetManager = contextWrapper.assets
			val inputStream = assetManager.open(assetName)
			val data = ByteArray(50)
			var current = 0
			while (inputStream.read(data, 0, data.size)
					.also { current = it } != -1) {
				buffer.write(data, 0, current)
			}
		} catch (e: Exception) {
			e.printStackTrace()
		}
		return buffer.toByteArray()
	}

	override fun emvProcessCallback(p0: ByteArray?) {
		var emvRetStatus: Byte? = p0?.get(0)
		var emvRetCode: Byte? = p0?.get(1)
		Log.i(TAG,"emvProcessCallback[$emvRetStatus][$emvRetCode]")
		uiUpdate?.addContentLog("emvProcessCallback[$emvRetStatus][$emvRetCode]")
		if(emvRetStatus == EMVJNIInterfaceAir.STATUS_COMPLETION)
		{
			readCardData()
			val iconType = if (emvRetCode?.toInt() == EmvProcess.APPROVE_OFFLINE || emvRetCode?.toInt() == EmvProcess.APPROVE_ONLINE)
				QMUITipDialog.Builder.ICON_TYPE_SUCCESS else QMUITipDialog.Builder.ICON_TYPE_FAIL
			uiUpdate?.showTipDialog("EMV Completed", iconType, 2000)
			uiUpdate?.setUpBtnStart(true)
		}
		else if(emvRetStatus == EMVJNIInterfaceAir.STATUS_CONTINUE)
		{
			EMVJNIInterfaceAir.emv_process_next()
		}
		else{
			val iconType = QMUITipDialog.Builder.ICON_TYPE_FAIL
			uiUpdate?.showTipDialog("EMV Error $p0?.get(1)", iconType, 2000)
			uiUpdate?.setUpBtnStart(true)
		}

	}

	fun readCardData()
	{
		val tagData = ByteArray(100)
		var tagDataLength = 0

		val pan: String
		val expiry:String
		val cardholderName:String
		val track2Data:String
		val csn:Byte

		// Card Number
		if(EMVJNIInterfaceAir.emv_is_tag_present(0x5A) >= 0) {
			tagDataLength = EMVJNIInterfaceAir.emv_get_tag_data(0x5A, tagData, tagData.size)
			pan = StringUtil.toString(
				AppUtil.removeTailF(
						ByteUtil.bcdToAscii(tagData, 0, tagDataLength)
					)
				)

			addContentLog("PAN: " + pan)
		}

		// Expiry
		if(EMVJNIInterfaceAir.emv_is_tag_present(0x5F24) >= 0) {
			tagDataLength = EMVJNIInterfaceAir.emv_get_tag_data(0x5F24, tagData, tagData.size)
			expiry = StringUtil.toHexString(
					tagData,
					0,
					3,
					false
				).substring(0, 4)

			addContentLog("expiry: " + expiry)
		}
		// Cardholder name
		if(EMVJNIInterfaceAir.emv_is_tag_present(0x5F20) >= 0) {
			tagDataLength = EMVJNIInterfaceAir.emv_get_tag_data(0x5F20, tagData, tagData.size)
			cardholderName = String(tagData, 0, tagDataLength)

			addContentLog("cardholderName: " + cardholderName)
		}
		// Track2
		if(emv_is_tag_present(0x57) >= 0) {
			tagDataLength = emv_get_tag_data(0x57, tagData, tagData.size)
			track2Data =
				StringUtil.toString(
					AppUtil.removeTailF(
						ByteUtil.bcdToAscii(tagData, 0, tagDataLength)
					)
				)

			addContentLog("track2Data: " + track2Data)
		}
		// CSN
		if(emv_is_tag_present(0x5F34) >= 0) {
			tagDataLength = emv_get_tag_data(0x5F34, tagData, tagData.size)
			csn = tagData[0]

			addContentLog("CSN: " + csn)
		}
	}


	override fun cardEventOccured(
		commandType: Int,
		command: ByteArray?,
		commandLen: Int,
		resp: ByteArray?,
		respSize: Int,
		respLen: ByteArray?
	): Int {

		//commandType: 0-card transmit , 1-offline pin,
		var tmpResp: ByteArray? = null
		try {
			if(App.instance.isoDep != null) {
				tmpResp = App.instance.isoDep!!.transceive(command!!.copyOf(commandLen))
			}
		} catch(e: Exception) {
			e.printStackTrace()
		}

		if(tmpResp != null && tmpResp.size <= respSize) {
			System.arraycopy(tmpResp, 0, resp, 0, tmpResp.size)
			System.arraycopy(NumberUtil.intToByte4LowFirst(tmpResp.size), 0, respLen, 0, 4)
			return 0
		}
		return -1
	}
}
